class MessageTextNotEnteredError(Exception):
    def __init__(self, message="Message text was not provided"):
        super().__init__(message)


class UnknownQuestionTypeError(Exception):
    def __init__(self, message="Unknown question type."):
        super().__init__(message)


class NoQuestionsEnteredError(Exception):
    def __init__(self, message="No questions were added."):
        super().__init__(message)

class QuestionNotFountError(Exception):
    def __init__(self, question_type:str):
        super().__init__(f"Тип вопроса {question_type} не поддерживается")


from pydantic import BaseModel, Field


class InfoButtons(BaseModel):
    finish_text:str = Field(default="Завершить",)
    multi_select_confirm:str = Field(default="Подтвердить выбор")
    start_again:str = Field(default="Начать сначала")
    confirm_entered_text:str = Field(default="Подтвердить введенные данные")
    class Config:
        from_attributes=True




import re
import phonenumbers
from phonenumbers.phonenumberutil import NumberParseException

def name(value: str) -> bool:
    """Проверяет, что имя содержит только буквы и дефис, длина 1-50 символов."""
    if not isinstance(value, str):
        return False
    if not (1 <= len(value) <= 50):
        return False
    return bool(re.match(r"^[A-Za-zА-Яа-яЁё\-]+$", value))

def phone_ru(value: str) -> bool:
    """Проверяет формат телефона +7XXXXXXXXXX или 8XXXXXXXXXX."""
    if not isinstance(value, str):
        return False
    pattern = r"^(?:\+7|8)\d{10}$"
    return bool(re.match(pattern, value))

def age(value: str) -> bool:
    """Возраст должен быть числом от 0 до 120."""
    if not value.isdigit():
        return False
    age = int(value)
    return 0 <= age <= 120

def height(value: str) -> bool:
    """Рост в сантиметрах: от 30 до 300."""
    try:
        h = float(value)
        return 30 <= h <= 300
    except ValueError:
        return False

def weight(value: str) -> bool:
    """Вес в кг: от 2 до 500."""
    try:
        w = float(value)
        return 2 <= w <= 500
    except ValueError:
        return False

def gender(value: str) -> bool:
    """
    Валидатор пола с поддержкой русского и английского языков.

    Допустимые варианты (регистр не важен):
    - Русские: "мужской", "м", "женский", "ж", "не указан", "другое", "не скажу"
    - Английские: "male", "m", "female", "f", "other", "not specified", "prefer not to say"

    Возвращает True, если значение совпадает с одним из допустимых, иначе False.
    """
    if not isinstance(value, str):
        return False

    value_normalized = value.strip().lower()

    valid_genders = {
        # Русские варианты
        "мужской", "м",
        "женский", "ж",
        "не указан", "другое", "не скажу",
        # Английские варианты
        "male", "m",
        "female", "f",
        "other", "not specified", "prefer not to say"
    }

    return value_normalized in valid_genders


def phone(value: str, region: str = None) -> bool:
    """
    Универсальная валидация телефонного номера для любых стран.

    Аргументы:
        value: строка, номер телефона в любом формате, включая международный (+...).
        region: код страны ISO alpha-2 (например, 'RU', 'US'), если номер без кода страны,
                для правильной обработки регионального формата.

    Возвращает:
        True, если номер валиден, иначе False.
    """
    if not isinstance(value, str) or not value.strip():
        return False
    try:
        # Разбор номера, учитывая регион, если он указан
        phone_number = phonenumbers.parse(value, region)
        # Проверка валидности номера
        return phonenumbers.is_valid_number(phone_number)
    except NumberParseException:
        return False





from aiogram.enums import ContentType
from aiogram.types import InlineKeyboardMarkup, ReplyKeyboardMarkup
from humanfriendly.terminal import message
from pydantic import  ValidationError
from aiogram import types, Dispatcher, F
from aiogram.fsm.context import FSMContext
from aiogram_dialog import Dialog, DialogManager, Window, StartMode, setup_dialogs
from aiogram_dialog.widgets.kbd import Button
from aiogram_dialog.widgets.text import Const
from aiogram_dialog.widgets.input import MessageInput
from aiogram_dialog.widgets.media import StaticMedia
from aiogram.filters import Command
from aiogram.fsm.state import StatesGroup, State
from pydantic import BaseModel, create_model, Field
from typing import Optional, Callable, Set, Generic, TypeVar, Type, Dict, Tuple, Any, List

from .builders.questions import QuestionBuilder
from .exceptions.questions import NoQuestionsEnteredError, MessageTextNotEnteredError, QuestionNotFountError
from .exceptions.validators import ValidatorNotFountError, EmptyValidatorNameError
from .models.buttons import InfoButtons
from .models.messages import InfoMessages
from .models.question import Question, QuestionType


from .utils.next_question_switcher_decorator import auto_switch_next_question
from ..utils import find_validator_by_name

ResultModelType = TypeVar("ResultModelType", bound=BaseModel)

class BriefSurvey(Generic[ResultModelType]):
    """
        Универсальный динамический опросник для Telegram-ботов на базе aiogram_dialog.

        Позволяет быстро создавать диалоговые опросы с любым числом вопросов и разными типами ответов.
        Вопросы можно добавлять как при инициализации, так и динамически через метод `add_question`.

        Атрибуты:
            questions (List[Question]): Список вопросов опроса.
            save_handler (Callable): Асинхронная функция для сохранения результата опроса.
            result_model (BaseModel): Pydantic-модель для итоговых данных.
            command_start (str): Команда для запуска опроса.
            info_messages (InfoMessages): Сообщения для пользователя (ошибки, подсказки).
            buttons (InfoButtons): Тексты кнопок для управления опросом.
            states_prefix (str): Префикс для состояний FSM.

        Примеры:

        from brief_survey import BriefSurvey
        from pydantic import BaseModel
        from typing import Optional


        class SurveyResult(BaseModel):
            name: Optional[str]
            gender: Optional[str]

        async def save_handler(user_id: int, result:SurveyResult | any):
            #динамическое обращение к полям результата опроса по имени вопроса если не указана модель.
            name = result.mame
            age = result.age
            gender = result.gender
            return
        survey = BriefSurvey(
            save_handler=save_handler,
            result_model=SurveyResult, # Опиционально
            start_command='start_brief' # Можно настраивать команду начала опроса
        )

        """

    def __init__(
            self,
            *,
            save_handler: Callable[[int, Any], Any],
            result_model: Optional[Type[ResultModelType]] = None,
            questions: Optional[List[Question]] = None,
            states_prefix: str = "SurveyStates",
            start_command: str = "start_survey",
            final_reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup = None
    ):
        """
               Инициализация опросника.

               Example:
               from brief_survey import BriefSurvey
               from pydantic import BaseModel
               from typing import Optional


               class SurveyResult(BaseModel):
                   name: Optional[str]
                    gender: Optional[str]


               #динамическое обращение к полям результата опроса по имени вопроса если не указана модель
               async def save_handler(user_id: int, result:SurveyResult | any):
                    name = result.mame
                    age = result.age
                    gender = result.gender
                    return


               survey = BriefSurvey(
                    save_handler=save_handler,
                    result_model=SurveyResult, #опционально
                    start_command='start_brief' #Можно настраивать команду начала опроса
                )

               Args:
                   save_handler: Асинхронная функция для сохранения результата (user_id, result).
                   result_model: Pydantic-модель результата (опционально, может быть создана автоматически).
                   questions: Список вопросов (можно добавить позже через add_question).
                   states_prefix: Префикс для состояний FSM.
                   start_command: Команда для запуска опроса в Telegram.
               """
        self._questions = questions
        self.save_handler = save_handler
        self.result_model = result_model
        self.command_start = start_command
        self.info_messages: InfoMessages = InfoMessages()
        self.buttons: InfoButtons = InfoButtons()
        self.states_prefix = states_prefix
        self.final_reply_markup = final_reply_markup
        self._dialog = None
        self.state: FSMContext = None

    @staticmethod
    def _create_states_group(class_name: str, state_names: List[str]):
        attrs = {name: State() for name in state_names}
        group = type(class_name, (StatesGroup,), attrs)
        mapping = {name: getattr(group, name) for name in state_names}
        return group, mapping

    @property
    def questions(self):
        return self._questions

    @property
    def dialog(self) -> Dialog:
        return self._dialog

    @auto_switch_next_question
    async def _process_text_input(self, message: types.Message,dialog: Dialog, manager: DialogManager):
        text = message.text.strip()
        state_name = manager.current_context().state.state.split(":")[1]
        question = self._get_question(state_name)
        if not question:
            await message.answer(self.info_messages.question_not_found)
            return

        if question.validator and not question.validator(text):
            if question.validator_error_message:
                error_text = question.validator_error_message
            else:
                error_text = self.info_messages.invalid_input
            await message.answer(error_text)
            return True

        if question.forced_exit_validator and not question.forced_exit_validator(text):
            await self.forced_exit_on_validation_error_handler(message, manager)
            return True

        manager.current_context().dialog_data[question.name] = text

    @auto_switch_next_question
    async def _process_text_input_with_confirmation(self, message: types.Message, dialog: Dialog, manager: DialogManager):
        text = message.text.strip()
        state_name = manager.current_context().state.state.split(":")[1]

        question = self._get_question(state_name)
        if not question:
            await message.answer(self.info_messages.question_not_found)
            return
        ctx_data = manager.current_context().dialog_data

        ctx_data[f"with_confirm_{state_name}"] = text


        if question.validator and not question.validator(text):
            if not question.validator_error_message:
                error_text = self.info_messages.invalid_input
            else:
                error_text = question.validator_error_message
            await message.answer(error_text)
            return True

        manager.current_context().dialog_data[question.name] = text
        text = f"Подтвердите введенные данные или отправьте данные заново.\n{question.confirm_field_name} {text}"
        try:
            # await message.delete()
            await message.answer(text)

        except Exception as ex:
            print(ex)

    @auto_switch_next_question
    async def _process_choice_selected(self, c: types.CallbackQuery, widget: Button, manager: DialogManager):
        # selected = widget.widget_id
        selected = widget.text.text  # Получаем текст кнопки, а не id (callback_data)
        state_name = manager.current_context().state.state.split(":")[1]

        question = self._get_question(state_name)
        if not question:
            await c.answer(self.info_messages.question_not_found)
            return True

        manager.current_context().dialog_data[question.name] = selected
        await c.answer()

    @auto_switch_next_question
    async def _process_multi_choice_selected(self, c: types.CallbackQuery, widget: Button, manager: DialogManager):
        selected_text = widget.text.text
        state_name = manager.current_context().state.state.split(":")[1]
        question = self._get_question(state_name)
        if not question:
            await c.answer(self.info_messages.question_not_found)
            return True

        ctx_data = manager.current_context().dialog_data
        multi_selected = ctx_data.get(f"multi_selected_{state_name}", set())
        if not isinstance(multi_selected, set):
            multi_selected = set(multi_selected)

        if selected_text in multi_selected:
            multi_selected.remove(selected_text)
        else:
            multi_selected.add(selected_text)
        ctx_data[f"multi_selected_{state_name}"] = multi_selected

        ctx_data[question.name] = ", ".join(multi_selected)
        try:

            await c.answer(f"Выбрано: {', '.join(multi_selected) if multi_selected else 'ничего'}")
        except:
            pass

    @auto_switch_next_question
    async def _process_media_input(self, message: types.Message, dialog: Dialog, manager: DialogManager):
        state_name = manager.current_context().state.state.split(":")[1]
        question = self._get_question(state_name)
        if not question:
            await message.answer(self.info_messages.question_not_found)
            return True

        if message.photo:
            file_id = message.photo[-1].file_id
        elif message.video:
            file_id = message.video.file_id
        else:
            await message.answer(self.info_messages.invalid_input)
            return True

        ctx_data = manager.current_context().dialog_data
        media_list = ctx_data.get(question.name, None)
        # if not isinstance(media_list, list):
        #     media_list = [media_list]
        # media_list.append(file_id)

        ctx_data[question.name] = file_id

    @auto_switch_next_question
    async def _process_media_list_input(self, message: types.Message, dialog: Dialog, manager: DialogManager):
        state_name = manager.current_context().state.state.split(":")[1]
        question = self._get_question(state_name)
        if not question:
            await message.answer(self.info_messages.question_not_found)
            return True

        if message.photo:
            file_id = message.photo[-1].file_id
        elif message.video:
            file_id = message.video.file_id
        else:
            await message.answer(self.info_messages.invalid_input)
            return True

        ctx_data = manager.current_context().dialog_data
        media_list = ctx_data.get(question.name, None)
        # if not isinstance(media_list, list):
        #     media_list = [media_list]
        media_list.append(file_id)
        ctx_data[question.name] = media_list
        if question.next_question:
            await manager.switch_to(self.state_map[question.next_question])
        else:
            await manager.next()
        return

    @auto_switch_next_question
    async def _confirm_multi_choice(self, c: types.CallbackQuery, widget: Button, manager: DialogManager):
        ctx_data = manager.current_context().dialog_data

        state_name = manager.current_context().state.state.split(":")[1]
        multi_selected = ctx_data.get(f"multi_selected_{state_name}", set())
        question = self._get_question(state_name)

        ctx_data[question.name] = ", ".join(multi_selected)
        await c.answer()

    @auto_switch_next_question
    async def _confirm_text_with_confirmation(self, c: types.CallbackQuery, button: Button, manager: DialogManager,*args,**kwargs):
        ctx_data = manager.current_context().dialog_data

        state_name = manager.current_context().state.state.split(":")[1]
        text = ctx_data.get(f"with_confirm_{state_name}", "")
        if not text:
            await c.answer(self.info_messages.no_confirmed_data)
            return
        question = self._get_question(state_name)
        if question.forced_exit_validator and not question.forced_exit_validator(text):
            await self.forced_exit_on_validation_error_handler(c.message, manager)
            return True
        ctx_data[question.name] = text
        await c.answer()

    async def forced_exit_on_validation_error_handler(self, message: types.Message, manager: DialogManager):
        await message.answer(self.info_messages.forced_exit_message)
        await manager.done()

    def _get_question(self, name: str) -> Optional[Question]:

        for q in self.questions:
            if q.name == name:
                return q
        return None

    def _make_window_for_question(self, question: Question) -> Window:
        state = self.state_map[question.name]
        qtext = question.text

        # Базовые элементы окна
        elements = []
        if qtext:
            elements.append(Const(qtext))
        if question.media:
            elements.append(StaticMedia(path=question.media))

        # Обработка по типу вопроса
        if question.type in ["text", "number"]:
            elements.append(MessageInput(self._process_text_input))
        elif question.type =="with_confirm":
            elements.append(MessageInput(self._process_text_input_with_confirmation))

            confirm_btn = Button(Const(f"Подтвердить {question.confirm_field_name.replace(':','')}"
                if question.confirm_field_name  else self.buttons.confirm_entered_text), id="confirm_text",
                                 on_click=self._confirm_text_with_confirmation)
            elements.append(confirm_btn)
        elif question.type == "choice":
            buttons = [
                Button(text=Const(label), id=key, on_click=self._process_choice_selected)
                for key, label in question.choices  # type: ignore
            ]
            elements.extend(buttons)
        elif question.type == "multi_choice":
            buttons = [
                Button(text=Const(label), id=str(i), on_click=self._process_multi_choice_selected)
                for i, (_, label) in enumerate(question.choices)  # type: ignore
            ]
            confirm_btn = Button(
                Const(self.buttons.multi_select_confirm  ),
                                 id="confirm",
                                 on_click=self._confirm_multi_choice)
            elements.extend(buttons)
            elements.append(confirm_btn)
        elif question.type in ["photo", "video", "media"]:
            if question.type == "photo":
                allowed_types = [ContentType.PHOTO]
            elif question.type == "video":
                allowed_types = [ContentType.VIDEO]
            else:
                allowed_types = [ContentType.PHOTO, ContentType.VIDEO]
            elements.append(MessageInput(self._process_media_input, content_types=allowed_types))
        else:
            raise QuestionNotFountError(question.type)

        return Window(*elements, state=state)

    async def _on_finish(self, c: types.CallbackQuery, button, manager: DialogManager):
        data = manager.current_context().dialog_data
        user_id = c.from_user.id
        try:
            result_obj = self.result_model.parse_obj(data)
        except ValidationError as e:
            await c.message.answer(f"Некорректные данные:\n" + "\n".join(
                [f"{err['loc'][0]}: {err['msg']}" for err in e.errors()]
            ))
            return True

        mes = await c.message.answer(self.info_messages.pre_save_message)
        markup = None
        if self.final_reply_markup:
            markup = self.final_reply_markup
        try:
            await self.save_handler(user_id, result_obj)
        except Exception as e:
            print("Save handler error:", e)
            await c.message.answer(self.info_messages.save_fail, reply_markup=markup)
        else:
            await c.message.answer(self.info_messages.save_success, reply_markup=markup)
        finally:
            await mes.delete()
            await c.message.delete()
            await manager.done()

    async def start(self, message: types.Message, dialog_manager: DialogManager, state: FSMContext):
        first_state_name = self.questions[0].name if self.questions else None
        if self.info_messages.start_message:
            await message.answer(self.info_messages.start_message)
        if first_state_name:
            self.state = state
            await state.set_state(self.state_map[first_state_name])
            await dialog_manager.start(self.state_map[first_state_name], mode=StartMode.RESET_STACK)
        else:
            await message.answer("Извините, еще нет вопросов для опроса.")

    async def _start_again(self, c: types.CallbackQuery, button, manager: DialogManager):
        first_state_name = self.questions[0].name if self.questions else None
        if first_state_name:
            await self.state.set_state(self.state_map[first_state_name])
            await manager.start(self.state_map[first_state_name], mode=StartMode.RESET_STACK)
        else:
            await c.message.answer("Извините, еще нет вопросов для опроса.")

    async def cmd_start_survey_handler(
            self,
            message: types.Message,
            dialog_manager: DialogManager,
            state: FSMContext,
    ):
        await self.start(message, dialog_manager, state)

    def register_handlers(self, dp: Dispatcher,
                          command_start: str = None,
                          callback_data: str = None,
                          text: str = None,
                          ):
        """
        Регистрирует все необходимые хендлеры для запуска и работы опроса в Telegram-боте.

        Args:
            dp (Dispatcher): Диспетчер aiogram.
            command_start (str, optional): Команда для запуска опроса.
            callback_data (str, optional): Callback data для запуска опроса по кнопке.
            text (str, optional): Текст сообщения для запуска опроса.
        """
        if not self.questions:
            raise NoQuestionsEnteredError
        else:
            self.create_result_model_from_questions()
        if command_start:
            dp.message.register(self.cmd_start_survey_handler, Command(command_start))
        else:
            dp.message.register(self.cmd_start_survey_handler, Command(self.command_start))
        if callback_data:
            dp.callback_query.register(self.cmd_start_survey_handler, F.data == callback_data)
        if text:
            dp.message.register(self.cmd_start_survey_handler, F.text == text)
        self.States, self.state_map = self._create_states_group(
            class_name=self.states_prefix,
            state_names=[q.name for q in self.questions] + ["finish"],
        )

        self.windows = [self._make_window_for_question(q) for q in self.questions]
        self.windows.append(
            Window(
                Const(self.info_messages.finish_text),

                Button(Const(self.buttons.finish_text), id="finish", on_click=self._on_finish),
                Button(Const(self.buttons.start_again), id="start_again", on_click=self._start_again),
                state=self.state_map["finish"],
            )
        )

        self._dialog = Dialog(*self.windows)
        dp.include_router(self.dialog)
        setup_dialogs(dp)

    def add_question(
            self,
            text: str,
            question_type: QuestionType = "text",
            name: str = None,
            choices: Optional[List[str] | Tuple[str] | Set[str]] = None,
            validator: Optional[Callable[[str], bool]|str] = None,
            next_questions: Optional[Dict[str, str]] = None,  # например {"Yes": "q3", "No": "q4"},
            next_question: Optional[str] = None,  # name следующего вопроса, нужно для ветвления запросов
            media_path: Optional[str] = None,
            forced_exit_validator: Optional[Callable[[str], bool]] = None,
            validate_by_question_name:bool=True,
            validator_error_message:Optional[str]=None,
            confirm_field_name:Optional[str]=None,
            *args,
            **kwargs
    ) -> Question:
        """
               Добавляет новый вопрос в опросник.

               Args:
                   text (str): Текст вопроса (обязателен).
                   question_type (str): Тип вопроса: "text", "number", "choice", "multi_choice", "photo", "video", "media".
                   name (str, optional): Уникальное имя вопроса (если не указано — генерируется автоматически).
                   choices (list, optional): Список вариантов для "choice" и "multi_choice".
                   validator (Callable, optional): Функция-валидатор для ответа.
                   next_questions (dict, optional): Словарь переходов к следующим вопросам по ответу.
                   next_question (str, optional): Имя следующего вопроса (для линейного перехода).
                   media_path (str, optional): Путь к медиафайлу (картинка, видео).
                   *args: Дополнительные позиционные аргументы.
                   **kwargs: Дополнительные именованные аргументы.

               Raises:
                   MessageTextNotEnteredError: Если текст вопроса пустой.

               Пример:
                   survey.add_question(
                       text="Ваш возраст?",
                       question_type="number",
                       name="age",
                       validator=lambda x: x.isdigit() and 0 < int(x) < 120
                   )
                   :param text:
                   :param question_type:
                   :param name:
                   :param choices:
                   :param validator:
                   :param next_questions:
                   :param next_question:
                   :param media_path:
                   :param validator_error_message:
                   :param validate_by_question_name:
                   :param forced_exit_validator:
               """

        if not text:
            raise MessageTextNotEnteredError("Текст вопроса не может быть пустым.")
        if choices:
            choices = [(str(i[0]), i[1]) for i in enumerate(choices)]
        if choices and question_type == "text":
            question_type = 'choice'
        if not name:
            name = f"q{len(self.questions) + 1}"

        if type(validator) == str:
            find_validator = find_validator_by_name(validator)
            if not find_validator:
                raise ValidatorNotFountError(validator)
            validator = find_validator
        if not validator and validate_by_question_name:
            find_validator = find_validator_by_name(name)
            validator = find_validator
        question_model = QuestionBuilder.create(
            text=text,
            question_type=question_type,
            name=name,
            choices=choices,
            validator=validator,
            next_questions=next_questions,
            next_question=next_question,
            media=media_path,
            forced_exit_validator=forced_exit_validator,
            validator_error_message=validator_error_message,
            confirm_field_name=confirm_field_name,
            *args,
            **kwargs
        )
        if not self.questions:
            self._questions = []
        self._questions.append(question_model)
        return question_model

    @staticmethod
    def get_field_type_and_default(question_type: str) -> Tuple[Any, Any]:

        if question_type in ("text", "choice", "multi_choice"):
            return (str, Field(default=None))
        elif question_type == "number":
            return (float, Field(default=None))
        elif question_type in ("photo", "video"):
            return (Optional[str], Field(default=None))
        elif question_type in ("media"):
            return (Optional[List[str]], Field(default_factory=list))
        else:
            return (str, Field(default=None))

    def create_result_model_from_questions(self) -> BaseModel:
        """
                Автоматически создает Pydantic-модель результата на основе списка вопросов.

                Returns:
                    BaseModel: Класс модели результата для сериализации итоговых данных пользователя.
                """
        fields: Dict[str, Tuple[Any, Any]] = {}

        for q in self.questions:
            field_type, default = self.get_field_type_and_default(q.type)
            fields[q.name] = (field_type, default)
        if not self.result_model:
            self.result_model = create_model('DynamicResultModel', **fields)
            return self.result_model
        else:
            model = create_model('DynamicResultModel', **fields)
            return model





from .core.models.question import QuestionBase, ChoiceQuestion, MultiChoiceQuestion, SurveyResult

from .core.survey import BriefSurvey

__all__= [
    "validators",
    "BriefSurvey",
    'QuestionBase',
    'ChoiceQuestion',
    'MultiChoiceQuestion',
]


import re

def validate_date_iso(value: str) -> bool:
    """Проверка формата даты ISO 8601 YYYY-MM-DD."""
    pattern = r"^\d{4}-\d{2}-\d{2}$"
    return bool(re.match(pattern, value))

def validate_date_flexible(value: str) -> bool:
    """
    Проверяет дату, где допустимы разделители '.', '-', ' '.

    Форматы, которые валидирует:
    - YYYY-MM-DD
    - DD-MM-YYYY
    - YYYY.MM.DD
    - DD.MM.YYYY
    - YYYY MM DD
    - DD MM YYYY

    Возвращает True, если строка соответствует одному из вариантов.
    НЕ проверяет корректность даты (например, 31.02.2023 пройдет).
    """
    # Общий паттерн для триада чисел с разделителем однотипным по всей строке
    pattern = r"""
        ^                 # начало строки
        (?P<part1>\d{2,4})
        (?P<sep>[-. ])     # разделитель: - или . или пробел
        (?P<part2>\d{1,2})
        (?P=sep)           # тот же разделитель
        (?P<part3>\d{1,4})
        $                 # конец строки
    """

    match = re.match(pattern, value, re.VERBOSE)
    if not match:
        return False

    part1 = match.group("part1")
    part2 = match.group("part2")
    part3 = match.group("part3")

    # Определяем, какой формат — если первый или третий часть длиной 4 — год вероятно там
    # Поддерживаем форматы YYYY-MM-DD и DD-MM-YYYY, где год — 4 цифры, день/месяц — 1-2
    if len(part1) == 4 and len(part3) in (1, 2):
        year = part1
        month = part2
        day = part3
    elif len(part3) == 4 and len(part1) in (1, 2):
        day = part1
        month = part2
        year = part3
    else:
        # не удалось определить порядок с годом
        return False

    # Дополнительная базовая проверка чисел и диапазонов (упрощённо)
    try:
        y = int(year)
        m = int(month)
        d = int(day)
        if not (1 <= m <= 12):
            return False
        if not (1 <= d <= 31):
            return False
        if y < 1000 or y > 9999:
            return False
    except ValueError:
        return False

    return True


from typing import Optional, List, Callable, Dict

from brief_survey.core.models.question import QuestionBase
from brief_survey.core.exceptions.questions import UnknownQuestionTypeError
from brief_survey.core.models.question import QUESTION_TYPE_MAP


class QuestionBuilder:
    def __init__(self, question: QuestionBase):
        self.question = question

    @staticmethod
    def create(question_type: str,
               name: str,
               text: str,
               choices: Optional[List[tuple]] = None,
               validator:Callable=None,
               next_questions: Optional[Dict[str, str]] = None,
               next_question: Optional[str] = None,  # name следующего вопроса, нужно для ветвления запросов
               media: Optional[str] = None,
               validator_error_message:Optional[str] = None,
               confirm_field_name:Optional[str]=None,
               *args,
               **kwargs) -> 'QuestionBase':
        model_cls = QUESTION_TYPE_MAP.get(question_type)
        if not model_cls:
            raise UnknownQuestionTypeError
        return model_cls(name=name,
                         text=text,
                         type=question_type,
                         choices=choices,
                         validator=validator,
                         next_questions=next_questions,
                         next_question=next_question,
                         media=media,
                         validator_error_message=validator_error_message,
                         confirm_field_name=confirm_field_name,
                         *args,
                         **kwargs
                         )



import re

def validate_not_empty(value: str) -> bool:
    """Проверяет, что строка не пустая."""
    return bool(value and value.strip())

def validate_email(value: str) -> bool:
    """Простая проверка email через регулярное выражение."""
    pattern = r"^[\w\.-]+@[\w\.-]+\.\w+$"
    return bool(re.match(pattern, value))

def validate_zip_code(value: str) -> bool:
    """Проверяет почтовый индекс: 5 или 6 цифр подряд."""
    return bool(re.match(r"^\d{5,6}$", value))

def validate_username(value: str) -> bool:
    """Имя пользователя: только буквы, цифры и подчеркивания, длина 3-30 символов."""
    return bool(re.match(r"^\w{3,30}$", value))

def validate_positive_int(value: str) -> bool:
    """Проверяет, что значение - положительное целое число."""
    return value.isdigit() and int(value) > 0

def validate_url(value: str) -> bool:
    """Простая проверка URL."""
    pattern = r"^(https?://)?([\w\-]+\.)+[\w\-]+(/[\w\-./?%&=]*)?$"
    return bool(re.match(pattern, value))

def validate_password_strength(value: str) -> bool:
    """
    Проверка пароля:
    - минимум 8 символов,
    - есть хотя бы одна цифра,
    - есть хотя бы одна заглавная буква,
    - есть хотя бы одна строчная буква,
    - есть хотя бы один спецсимвол.
    """
    if len(value) < 8:
        return False
    if not re.search(r"\d", value):
        return False
    if not re.search(r"[A-Z]", value):
        return False
    if not re.search(r"[a-z]", value):
        return False
    if not re.search(r"[!@#$%^&*(),.?\":{}|<>]", value):
        return False
    return True




from typing import List, Optional, Tuple, Callable, Union, Literal, Dict
from pydantic import BaseModel, field_validator

from brief_survey.core.exceptions.questions import UnknownQuestionTypeError

QuestionType = Literal["text", "number",'with_confirm', "choice", "multi_choice", "photo", "video", "media"]


class QuestionBase(BaseModel):
    name: str
    text: str
    type: QuestionType
    validator: Optional[Callable[[str], bool]] = None
    next_questions: Optional[Dict[str, str]] = None  # например {"Yes": "q3", "No": "q4"}
    next_question: Optional[str] = None,  # name следующего вопроса, нужно для ветвления запросов
    media: Optional[str] = None,
    forced_exit_validator: Optional[Callable[[str], bool]] = None
    validator_error_message: Optional[str] =None
    confirm_field_name:Optional[str] ="Введенные данные:"
    @field_validator('type')
    def type_must_be_known(cls, v):
        if v not in QuestionType.__args__:  # type: ignore
            raise UnknownQuestionTypeError
        return v


class ChoiceQuestion(QuestionBase):
    choices: List[Tuple[str | int, str]]

    @field_validator("choices")
    def check_choices_non_empty(cls, v, values):
        if not v or not isinstance(v, list):
            raise ValueError("Choices must be a non-empty list")
        return v


class MultiChoiceQuestion(QuestionBase):
    choices: List[Tuple[str, str]]

    @field_validator("choices")
    def check_choices_non_empty(cls, v, values):
        if not v or not isinstance(v, list):
            raise ValueError("Choices must be a non-empty list")
        return v


Question = Union[QuestionBase, ChoiceQuestion, MultiChoiceQuestion]


# Пример модели результата, пользователь сам может создать свою
class SurveyResult(BaseModel):
    name: Optional[str]
    age: Optional[int]
    gender: Optional[str]


QUESTION_TYPE_MAP = {
    "text": QuestionBase,
    "with_confirm": QuestionBase,
    "photo": QuestionBase,
    "video": QuestionBase,
    "media": QuestionBase,
    "number": QuestionBase,
    "choice": ChoiceQuestion,
    "multi_choice": MultiChoiceQuestion,
}


from pydantic import BaseModel, Field


class InfoMessages(BaseModel):
    invalid_input:str = Field(default="Пожалуйста, введите корректные данные.",)
    save_success:str = Field(default="Спасибо! Данные успешно сохранены.",)
    save_fail:str = Field(default="Произошла ошибка при сохранении. Попробуйте позже.",)
    finish_text:str = Field(default="Данные приняты.",)
    question_not_found:str = Field(default= "Ошибка: вопрос не найден.")
    pre_save_message:str = Field(default= "Сохраняю")
    start_message:str = Field(default=None)
    forced_exit_message:str = Field(default="Выход из опроса.Введенные данные не позволяют продолжить опрос")
    no_confirmed_data:str = Field(default='Укажите запрашиваемые данные чтобы продолжить...')

    class Config:
        from_attributes=True



from typing import List

from brief_survey import QuestionBase
from brief_survey.core.db.repositories.base import BaseRepository, T


# class CreditsRepository(BaseRepository[Credits]):
#     def __init__(self):
#         super().__init__(
#             sql_model_class=Credits,
#             async_mode=True,
#             connection_string=async_url,
#         )

class CompleteSurveyManager:
    def __init__(self,
                 repository:BaseRepository[T]=None,
                 db_tablename:str="Survey"
                 ):
        self.repo = None

    async def asave_survey(self,name:str,questions:List[QuestionBase]):
        ...

    def save_survey(self,name:str,questions:List[QuestionBase]):
        ...

    async def aget_survey_by_name(self,name:str):
        ...

    def get_survey_by_name(self,name:str):
        ...

    async def aget_survey_by_id(self,idx:str|int):
        ...

    def get_survey_by_id(self,idx:str|int):
        ...


from functools import wraps
from aiogram_dialog.manager.manager import ManagerImpl
from aiogram_dialog.widgets.kbd import Button

def auto_switch_next_question(func):
    """
    Декоратор для методов класса, который после выполнения
    переключит состояние в диалоге, используя self из метода.
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        result = await func(self, *args, **kwargs)
        if result:
            return
        manager = kwargs.get("manager", None)
        widget = kwargs.get("widget", None)
        for arg in args:
            if isinstance(arg, ManagerImpl):
                manager = arg
            if isinstance(arg, Button):
                widget = arg
        if not manager:
            return result

        state_name = manager.current_context().state.state.split(":")[1]
        question = self._get_question(state_name)

        if question:
            next_state_name = None
            selected = manager.current_context().dialog_data.get(question.name, None)
            if question.next_questions and selected in question.next_questions:
                next_state_name = question.next_questions[selected]
            elif question.next_question:
                next_state_name = question.next_question

            if question.type == 'with_confirm' and widget and widget.widget_id == "confirm_text":
                if not selected:
                    return result
                if next_state_name:
                    await manager.switch_to(self.state_map[next_state_name])
                else:
                    await manager.next()
            elif question.type == 'with_confirm' and selected:
                return result
            elif question.type == 'multi_choice' and widget.widget_id == "confirm":
                    await manager.switch_to(self.state_map[next_state_name])
            elif question.type == 'multi_choice' and selected:
                return result
            elif next_state_name:
                await manager.switch_to(self.state_map[next_state_name])
            else:
                await manager.next()

        return result

    return wrapper




class ValidatorNotFountError(Exception):
    def __init__(self, validator_name:str):
        super().__init__(f"Validator {validator_name} not found")

class EmptyValidatorNameError(Exception):
    def __init__(self, validator_name:str):
        super().__init__(f"Empty validator name")


from setuptools import setup, find_packages

def readme():

  with open('README.md', 'r') as f:
    return f.read()

setup(
    name="brief_survey",
    version="0.2.6",
    description="Dynamic survey/dialog for aiogram3  with aiogram_dialog and Pydantic support",
    author="Fugguri",
    url="https://github.com/Fugguri/brief_survey",
    packages=find_packages(),
    long_description_content_type='text/markdown',
    long_description=readme(),
    classifiers=[
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    install_requires=[

        "aiogram>=3.20",
        "aiogram_dialog>=2.3.1",
        "phonenumbers>=9.0.10",
        "pydantic>=2.11.7",
        "humanfriendly>=10.0",
    ],
    keywords='aiogram3,aiogram, aiogram_dialog, brief  ',
    project_urls={
        'pypi': 'https://pypi.org/project/brief-survey/',
        'github':'https://github.com/Fugguri/brief_survey'
    },
    python_requires='>=3.10',

)



import os
import importlib
import inspect
import pkgutil
import brief_survey.validators  # ваш подпакет с валидаторами

def find_validator_by_name(name: str):
    # Получаем путь к пакету validators
    package = brief_survey.validators
    package_path = package.__path__  # список путей, обычно один

    for _, module_name, is_pkg in pkgutil.iter_modules(package_path):
        if not is_pkg:
            full_module_name = f"{package.__name__}.{module_name}"
            module = importlib.import_module(full_module_name)
            for func_name, func in inspect.getmembers(module, inspect.isfunction):
                if func_name == name:
                    return func
    return None




import datetime
import pytz
from sqlalchemy import (
    Column, Enum, Integer, String, DateTime, Boolean, BigInteger,
    ForeignKey, TEXT, Text, Float, UniqueConstraint, Date
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

def utcnow():
    return datetime.datetime.now(pytz.UTC)

Base = declarative_base()





import contextlib
import csv
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    cast, TypeVar, Generic,
)

import openpyxl
from sqlalchemy import delete, select, update
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    create_async_engine,
)
from sqlalchemy.engine import Engine
from sqlalchemy.orm import (
    Session as SQLSession,
    scoped_session,
    sessionmaker,
)

from openpyxl.utils import get_column_letter

try:
    from sqlalchemy.ext.asyncio import async_sessionmaker
except ImportError:
    async_sessionmaker = type("async_sessionmaker", (type,), {})  # type: ignore[assignment,misc]

DBConnection = Union[AsyncEngine, Engine, str]

from ..db_models import Base

T = TypeVar("T", bound=Base)

class BaseRepository(Generic[T]):
    """
    Универсальный репозиторий для SQLAlchemy моделей с произвольными CRUD операциями.

    Все фильтры для get/update/delete передаются через kwargs.
    Создание и обновление с передачей kwargs для полей.
    """

    @property
    def Session(self) -> Union[scoped_session, async_sessionmaker]:
        return self.session_maker

    def __init__(
        self,
        sql_model_class: Type[T],
        connection_string: Optional[str] = 'sqlite+aiosqlite:///brief.db',
        connection: Optional[DBConnection] = None,
        engine_args: Optional[Dict[str, Any]] = None,
        async_mode: Optional[bool] = None,
        table_name: Optional[str] = None,

    ):


        if connection_string and connection:
            raise ValueError("connection_string and connection are mutually exclusive")

        if connection_string:
            connection = connection_string



        if isinstance(connection, str):
            self.async_mode = async_mode if async_mode is not None else False
            if self.async_mode:
                self.async_engine = create_async_engine(connection, **(engine_args or {}))
            else:
                from sqlalchemy import create_engine

                self.engine = create_engine(connection, **(engine_args or {}))
        elif isinstance(connection, Engine):
            self.async_mode = False
            self.engine = connection
        elif isinstance(connection, AsyncEngine):
            self.async_mode = True
            self.async_engine = connection
        else:
            raise ValueError(
                "connection should be connection string or instance of "
                "sqlalchemy.engine.Engine or sqlalchemy.ext.asyncio.engine.AsyncEngine"
            )

        self.session_maker: Union[scoped_session, async_sessionmaker]
        if self.async_mode:
            self.session_maker = async_sessionmaker(bind=self.async_engine)
        else:
            self.session_maker = scoped_session(sessionmaker(bind=self.engine))
        if table_name:

            DynamicModel = type(
                f"{sql_model_class.__name__}DynamicTable",
                (sql_model_class,),
                {"__tablename__": table_name}
            )
            self.sql_model_class = DynamicModel
        else:
            self.sql_model_class = sql_model_class

        if not self.async_mode:
            self._create_table_if_not_exists()

        self._table_created = False

    def _create_table_if_not_exists(self) -> None:
        self.sql_model_class.metadata.create_all(self.engine)
        self._table_created = True

    async def _acreate_table_if_not_exists(self) -> None:
        if not self._table_created:
            assert self.async_mode, "Async only"
            async with self.async_engine.begin() as conn:
                await conn.run_sync(self.sql_model_class.metadata.create_all)
            self._table_created = True

    def get_all(self, *args: Any, **filters: Any) -> List[T]:
        """
            Получить все записи, соответствующие переданным фильтрам.

            Поддерживаемые фильтры:
                - Обычное равенство: field=value
                - "LIKE" с подстрокой: field__like=value

            Аргументы:
                *args: позиционные условия (обычно не используются).
                **filters: фильтры по полям модели.
                    Ключи с обычными именами фильтруют по равенству.
                    Ключи с суффиксом '__like' фильтруют по оператору LIKE.
                    Например:
                        get_all(status='active', title__like='пример')

            Возвращает:
                Список объектов модели, где все условия выполнены (AND).

            Исключения:
                AttributeError — при отсутствии поля в модели.
        """
        with self._make_sync_session() as session:
            query = session.query(self.sql_model_class)

            if filters:
                # Собираем условия
                conditions = []
                for key, value in filters.items():
                    if key.endswith("__like"):
                        field_name = key[:-6]
                        field = getattr(self.sql_model_class, field_name, None)
                        if field is None:
                            raise AttributeError(f"Model has no field '{field_name}'")
                        conditions.append(field.like(f"%{value}%"))
                    else:
                        field = getattr(self.sql_model_class, key, None)
                        if field is None:
                            raise AttributeError(f"Model has no field '{key}'")
                        conditions.append(field == value)
                query = query.filter(*conditions)

            query = query.order_by(self.sql_model_class.id.asc())
            return query.all()

    async def aget_all(self, *args: Any, **filters: Any) -> List[T]:
        """
        Асинхронно получить все записи, удовлетворяющие фильтрам.

        Поддерживаемые фильтры:
            - Обычное равенство: field=value
            - "LIKE" с подстрокой: field__like=value

        Аргументы:
            *args: позиционные условия (обычно не используются).
            **filters: фильтры по полям модели.
                Ключи с суффиксом '__like' используются для фильтрации через LIKE.
                Пример:
                    await aget_all(status='active', description__like='транспорт')

        Возвращает:
            Список объектов модели, удовлетворяющих всем фильтрам (логика AND).

        Исключения:
            AttributeError — если указано поле, отсутствующее в модели.
        """
        await self._acreate_table_if_not_exists()
        async with self._make_async_session() as session:
            stmt = select(self.sql_model_class)

            if filters:
                from sqlalchemy import and_
                conditions = []
                for key, value in filters.items():
                    if key.endswith("__like"):
                        field_name = key[:-6]
                        field = getattr(self.sql_model_class, field_name, None)
                        if field is None:
                            raise AttributeError(f"Model has no field '{field_name}'")
                        conditions.append(field.like(f"%{value}%"))
                    else:
                        field = getattr(self.sql_model_class, key, None)
                        if field is None:
                            raise AttributeError(f"Model has no field '{key}'")
                        conditions.append(field == value)
                stmt = stmt.where(and_(*conditions))

            stmt = stmt.order_by(self.sql_model_class.id.asc())
            result = await session.execute(stmt)
            return result.scalars().all()


    def get_one(self, **filters: Any) -> Optional[T]:
        """
        Получить одну запись по фильтрам.

        Возвращает None, если не найдено.
        """
        with self._make_sync_session() as session:
            query = session.query(self.sql_model_class)
            if filters:
                query = query.filter_by(**filters)
            res:T = query.first()
            return res

    async def aget_one(self, **filters: Any) -> Optional[T]:
        """Асинхронный аналог get_one."""
        await self._acreate_table_if_not_exists()
        async with self._make_async_session() as session:
            stmt = select(self.sql_model_class)
            if filters:
                stmt = stmt.filter_by(**filters)
            result = await session.execute(stmt)
            return result.scalars().first()

    def create(self, **fields: Any) -> T:
        """
        Создать новую запись с переданными полями.

        Возвращает созданный объект (но ещё не сохранён в БД, если требуется можно вызвать add_item).
        """
        obj = self.sql_model_class(**fields)
        self.add_item(obj)
        return obj

    async def acreate(self, **fields: Any) -> T:
        """
        Асинхронно создать новую запись и сохранить в БД.
        """
        obj = self.sql_model_class(**fields)
        await self.aadd_item(obj)
        return obj

    def update(self, values: Dict[str, Any], **filters: Any) -> int:
        """
        Обновить записи, удовлетворяющие фильтрам, новыми значениями.

        Args:
            values: словарь поле->значение для обновления.
            **filters: фильтры для выбора записей.

        Returns:
            Количество обновлённых записей.
        """
        with self._make_sync_session() as session:
            stmt = update(self.sql_model_class).where(
                *[getattr(self.sql_model_class, k) == v for k, v in filters.items()]
            ).values(**values)
            result = session.execute(stmt)
            session.commit()
            return result.rowcount or 0

    async def aupdate(self, values: Dict[str, Any], **filters: Any) -> int:
        """
        Асинхронное обновление записей по фильтрам.
        """
        await self._acreate_table_if_not_exists()
        async with self._make_async_session() as session:
            stmt = update(self.sql_model_class).where(
                *[getattr(self.sql_model_class, k) == v for k, v in filters.items()]
            ).values(**values)
            result = await session.execute(stmt)
            await session.commit()
            return result.rowcount or 0

    def delete(self, **filters: Any) -> int:
        """
        Удалить записи, удовлетворяющие фильтрам.

        Returns:
            Количество удалённых записей.
        """
        with self._make_sync_session() as session:
            stmt = delete(self.sql_model_class).where(
                *[getattr(self.sql_model_class, k) == v for k, v in filters.items()]
            )
            result = session.execute(stmt)
            session.commit()
            return result.rowcount or 0

    async def adelete(self, **filters: Any) -> int:
        """Асинхронное удаление по фильтрам."""
        await self._acreate_table_if_not_exists()
        async with self._make_async_session() as session:
            stmt = delete(self.sql_model_class).where(
                *[getattr(self.sql_model_class, k) == v for k, v in filters.items()]
            )
            result = await session.execute(stmt)
            await session.commit()
            return result.rowcount or 0

    def add_item(self, item: Any) -> None:
        """Сохранить один объект (sync)."""
        with self._make_sync_session() as session:
            session.add(item)
            session.commit()

    async def aadd_item(self, item: Any) -> None:
        """Сохранить один объект (async)."""
        await self._acreate_table_if_not_exists()
        async with self._make_async_session() as session:
            session.add(item)
            await session.commit()

    @contextlib.contextmanager
    def _make_sync_session(self) -> Generator[SQLSession, None, None]:
        if self.async_mode:
            raise ValueError(
                "sync session requested, but async_mode=True, use async methods"
            )
        with self.session_maker() as session:
            yield cast(SQLSession, session)

    @contextlib.asynccontextmanager
    async def _make_async_session(self) -> AsyncGenerator[AsyncSession, None]:
        if not self.async_mode:
            raise ValueError(
                "async session requested, but async_mode=False, use sync methods"
            )
        async with self.session_maker() as session:
            yield cast(AsyncSession, session)

    def export_to_csv(
            self,
            items: List[Type[T]],
            filepath: str,
            fields: Optional[List[str]] = None,
            delimiter: str = ",",
            encoding: str = "utf-8"
    ) -> None:
        """
        Экспортирует список ORM объектов или схем в CSV файл.

        Args:
            items: Список объектов self.sql_model_class (или Pydantic моделей)
            filepath: Путь к файлу для записи
            fields: Список полей для экспорта. Если None — берутся все колонки модели.
            delimiter: Разделитель. По умолчанию запятая.
            encoding: Кодировка файла.
        """
        if not items:
            raise ValueError("Нет данных для экспорта")

        if fields is None:
            # Получаем все столбцы SQLAlchemy модели (если это ORM объекты)
            if hasattr(self.sql_model_class, "__table__"):
                fields = [col.name for col in self.sql_model_class.__table__.columns]
            else:
                # Если объекты не ORM, пробуем взять атрибуты первого объекта
                fields = list(vars(items[0]).keys())

        with open(filepath, mode="w", encoding=encoding, newline="") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fields, delimiter=delimiter)
            writer.writeheader()
            for item in items:
                row = {}
                for f in fields:
                    val = getattr(item, f, None)
                    # Если поле — datetime/date, конвертируем в строку
                    if hasattr(val, "isoformat"):
                        val = val.isoformat()
                    row[f] = val
                writer.writerow(row)

        return filepath

    def export_to_excel(
            self,
            items: List[Type[T]],
            filepath: str,
            fields: Optional[List[str]] = None,
    ) -> str:
        """
        Экспортирует список ORM объектов или схем в Excel файл (xlsx).

        Args:
            items: Список объектов self.sql_model_class или Pydantic моделей
            filepath: Путь к результату, например 'export.xlsx'
            fields: Поля для экспорта. Если None — все столбцы модели.
        Return:
            str - filepath
        """
        if not items:
            raise ValueError("Нет данных для экспорта")

        if fields is None:
            if hasattr(self.sql_model_class, "__table__"):
                fields = [col.name for col in self.sql_model_class.__table__.columns]
            else:
                fields = list(vars(items[0]).keys())

        wb = openpyxl.Workbook()
        ws = wb.active
        ws.title = "Export"

        # Запишем заголовки
        for col_idx, field in enumerate(fields, start=1):
            cell = ws.cell(row=1, column=col_idx)
            cell.value = field

        # Запишем данные
        for row_idx, item in enumerate(items, start=2):
            for col_idx, field in enumerate(fields, start=1):
                val = getattr(item, field, None)
                if hasattr(val, "isoformat"):
                    val = val.isoformat()
                ws.cell(row=row_idx, column=col_idx, value=val)

        # Можно немного подстроить ширину столбцов
        for idx, col in enumerate(fields, start=1):
            max_length = max(
                len(str(getattr(item, col, "")) or "") for item in items
            )
            max_length = max(max_length, len(col))
            ws.column_dimensions[get_column_letter(idx)].width = max_length + 2

        wb.save(filepath)
        return filepath